<!doctype html>

<html>
	<head>
		<title>Psychology Experiment</title>
		<link rel=stylesheet href="/static/css/bootstrap.min.css" type="text/css">
		<link rel=stylesheet href="/static/css/style.css" type="text/css">
	</head>
	<body>
		<div id="container-ad">

			<div id="ad">
				<div class="row">
					<div class="col-xs-2">
						<!-- REPLACE THE LOGO HERE WITH YOUR  UNIVERSITY, LAB, or COMPANY -->
						<img id="adlogo" src="/static/images/university.png" alt="Lab Logo" />
					</div>
					<div class="col-xs-10">
						<h1>Thanks for your participation!</h1>
						<hr>
					    <div class="alert alert-warning">
					    	Our records suggest you have already completed this task.  Thank you.
					    	You cannot complete the same experiment twice.
					    </div>
					</div>
			</div>
		</div>
	</body>
</html>